import torch
import numpy as np
from torch.utils.data import Dataset
import enchanter.engine.modules as modules


def test_get_dataset_1():
    x = torch.randn(32, 128)
    y = torch.randint(0, 9, (32, ))

    ds = modules.get_dataset(x, y)
    return isinstance(ds, Dataset)


def test_get_dataset_2():
    x = torch.randn(32, 128)

    ds = modules.get_dataset(x)
    return isinstance(ds, Dataset)



from .positional_encoding import PositionalEncoding
from .mlp import MLP
from .mlp import PositionWiseFeedForward
from .se_layer import SELayer1d
from .se_layer import SELayer2d
from .dense_interpolation import DenseInterpolation
from .conv import CausalConv1d

__all__ = [
    "PositionalEncoding",
    "MLP",
    "PositionWiseFeedForward",
    "SELayer1d",
    "SELayer2d",
    "DenseInterpolation",
    "CausalConv1d"
]



import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader

import enchanter
from enchanter.callbacks import TensorBoardLogger

X = torch.randn(512, 10).numpy()
Y = torch.randint(0, 9, (512,)).numpy()


class Runner1(enchanter.engine.BaseRunner):
    def __init__(self):
        super(Runner1, self).__init__()
        self.model = nn.Linear(10, 10)
        self.optimizer = optim.Adam(self.model.parameters())
        self.experiment = TensorBoardLogger("../tmp")
        self.criterion = nn.CrossEntropyLoss()

    def train_step(self, batch):
        x, y = batch
        out = self.model(x)
        loss = self.criterion(out, y)

        return {"loss": loss}


class Runner2(enchanter.engine.BaseRunner):
    def __init__(self):
        super(Runner2, self).__init__()
        self.model = nn.Linear(10, 10)
        self.optimizer = optim.Adam(self.model.parameters())
        self.experiment = TensorBoardLogger("../tmp")
        self.criterion = nn.CrossEntropyLoss()

    def train_step(self, batch):
        x, y = batch
        out = self.model(x)
        loss = self.criterion(out, y)

        return {"loss": loss}

    def val_step(self, batch):
        x, y = batch
        out = self.model(x)
        loss = self.criterion(out, y)

        return {"loss": loss}

    def test_step(self, batch):
        x, y = batch
        out = self.model(x)
        loss = self.criterion(out, y)

        return {"loss": loss}


def test_custom_runner_1():
    runner = Runner1()
    ds = enchanter.engine.modules.get_dataset(X, Y)
    loader = DataLoader(ds, batch_size=32)

    try:
        runner.add_loader("train", loader)
        runner.train_config(epochs=1)
        runner.run()
        is_pass = True

    except Exception as e:
        print(e)
        is_pass = False

    assert is_pass


def test_custom_runner_2():
    runner = Runner2()
    ds = enchanter.engine.modules.get_dataset(X, Y)
    loader = DataLoader(ds, batch_size=32)

    try:
        runner.add_loader("test", loader)
        runner.train_config(epochs=1)
        runner.run()
        is_pass = True

    except Exception as e:
        print(e)
        is_pass = False

    assert is_pass


def test_custom_runner_3():
    runner = Runner2()

    ds = enchanter.engine.modules.get_dataset(X, Y)
    loader = DataLoader(ds, batch_size=32)

    try:
        runner.add_loader("train", loader)
        runner.add_loader("val", loader)
        runner.add_loader("test", loader)
        runner.train_config(epochs=1)
        runner.run()
        is_pass = True

    except Exception as e:
        print(e)
        is_pass = False

    assert is_pass



# ***************************************************
#  _____            _                 _
# | ____|_ __   ___| |__   __ _ _ __ | |_ ___ _ __
# |  _| | '_ \ / __| '_ \ / _` | '_ \| __/ _ \ '__|
# | |___| | | | (__| | | | (_| | | | | ||  __/ |
# |_____|_| |_|\___|_| |_|\__,_|_| |_|\__\___|_|
#
# ***************************************************

import re
import operator
from abc import ABC
from collections import OrderedDict
from typing import Any, Dict, Tuple, Union, List, Optional

from numpy import floor, ndarray
from tqdm.auto import tqdm
from torch import device
from torch.nn import Module
from torch.optim.optimizer import Optimizer
from torch.cuda import is_available
from torch.tensor import Tensor
from torch.autograd import no_grad
from torch.utils.data import DataLoader, SubsetRandomSampler
from comet_ml.experiment import BaseExperiment as BaseExperiment

from enchanter.engine.saving import RunnerIO
from enchanter.engine.modules import send, get_dataset
from enchanter.callbacks import BaseLogger as BaseLogger
from enchanter.callbacks import EarlyStopping as EarlyStopping


__all__ = [
    "BaseRunner"
]


class BaseRunner(ABC, RunnerIO):
    """
    PyTorchモデルの訓練に用いる Runner を作成する為のクラスです。


    Examples:

        >>> from comet_ml import Experiment
        >>> import torch
        >>> class Runner(BaseRunner):
        >>>     def __init__(self):
        >>>         super(Runner, self).__init__()
        >>>         self.model = torch.nn.Linear(10, 10)
        >>>         self.optimizer = torch.optim.Adam(self.model.parameters())
        >>>         self.experiment = Experiment()
        >>>         self.criterion = torch.nn.CrossEntropyLoss()
        >>>
        >>>     def train_step(self, batch):
        >>>         x, y = batch
        >>>         out = self.model(x)
        >>>         loss = self.criterion(out, y)
        >>>
        >>>         return {"loss": loss}

    """
    def __init__(self) -> None:
        super(BaseRunner, self).__init__()
        self.device: device = device("cuda" if is_available() else "cpu")
        self.model: Module = NotImplemented
        self.optimizer: Optimizer = NotImplemented
        self.scheduler: Optional[List] = None
        self.experiment: Union[BaseExperiment, BaseLogger] = NotImplemented
        self.early_stop: Optional[EarlyStopping] = None
        self.configures: Dict[str, Any] = {
            "epochs": 0
        }

        self.pbar = None
        self._loaders: Dict[str, DataLoader] = {}
        self._metrics: Dict = {}

    def backward(self, loss: Tensor) -> None:
        """
        calculate the gradient

        Warnings:
            This method may be changed from staticmethod in the future to support FP16 and others.

        Args:
            loss (torch.Tensor):

        Returns:
            None

        """
        loss.backward()

    def update_optimizer(self) -> None:
        self.optimizer.step()

    def update_scheduler(self, epoch: int) -> None:
        """
        スケジューラの値を更新するために呼ばれるメソッドです。
        Optimizerを更新した後に呼ばれます。

        Args:
            epoch (int): 現在のエポック数

        Returns:
            None

        Examples:
            >>> from torch.optim import lr_scheduler
            >>> from enchanter.wrappers import ClassificationRunner
            >>> runner: BaseRunner = ClassificationRunner(
            >>>     model=...,
            >>>     optimizer=...,
            >>>     criterion=...,
            >>>     scheduler=[
            >>>         lr_scheduler.CosineAnnealingLR(...), lr_scheduler.StepLR(...)
            >>>     ]
            >>> )

        """
        for scheduler in self.scheduler:
            scheduler.step()

        self.experiment.log_metric("scheduler_lr", self.scheduler[-1].get_last_lr(), epoch=epoch)

    def train_step(self, batch: Tuple) -> Dict[str, Tensor]:
        """
        ニューラルネットの訓練時、
            >>> import torch.nn as nn
            >>> train_loader: DataLoader = ...
            >>> model: nn.Module = ...
            >>> criterion = ...
            >>> for x, y in train_loader:
            >>>     out = model(x)
            >>>     loss = criterion(out, y)

        にあたる箇所を担当するメソッドです。

        Args:
            batch: PyTorch DataLoader から得られる訓練用のデータトラベルを含むタプル

        Returns:
            キーに 'loss' を含む辞書を返す必要があります。

        Examples:
            >>> def train_step(self, batch):
            >>>     x, y = batch
            >>>     out = self.model(x)
            >>>     loss = nn.functional.cross_entropy(out, y)
            >>>     return {"loss": loss}

        """

    def train_end(self, outputs: List) -> Dict[str, Tensor]:
        """
        ニューラルネットの訓練時、1step 終了ごとに実行されるメソッドです。

        Args:
            outputs:

        Returns:

        """
        return {}

    def val_step(self, batch: Tuple) -> Dict[str, Tensor]:
        """
        ニューラルネットの検証時、1step ごとに実行されるメソッドです。

        Args:
            batch: PyTorch DataLoader から得られる訓練用のデータトラベルを含むタプル

        Returns:
            辞書を返す必要があります。

        """

    def val_end(self, outputs: List) -> Dict[str, Tensor]:
        """
        ニューラルネットの検証時、1step 終了ごとに実行されるメソッドです。

        Args:
            outputs:

        Returns:

        """
        return {}

    def test_step(self, batch: Tuple) -> Dict[str, Tensor]:
        """
        ニューラルネットの評価時、1step ごとに実行されるメソッドです。

        Args:
            batch: PyTorch DataLoader から得られる訓練用のデータトラベルを含むタプル

        Returns:
            辞書を返す必要があります。

        """

    def test_end(self, outputs: List) -> Dict[str, Tensor]:
        """
        ニューラルネットの評価時、1step 終了ごとに実行されるメソッドです。

        Args:
            outputs:

        Returns:

        """
        return {}

    def train_cycle(self, epoch: int, loader: DataLoader) -> None:
        """
        ニューラルネットの訓練ループです。

        Args:
            epoch
            loader (torch.utils.data.DataLoader):

        """
        results = list()
        loader_size = len(loader)

        self.model.train()
        with self.experiment.train():
            for step, batch in enumerate(loader):
                self.optimizer.zero_grad()
                batch = send(batch, self.device)
                # on_step_start()
                outputs = self.train_step(batch)
                self.backward(outputs["loss"])
                self.update_optimizer()

                if hasattr(self.pbar, "set_postfix"):
                    per = "{:1.0%}".format(step / loader_size)
                    self.pbar.set_postfix(
                        OrderedDict(train_batch=per), refresh=True
                    )

                outputs = {
                    key: outputs[key].detach().cpu() if isinstance(outputs[key], Tensor) else outputs[key]
                    for key in outputs.keys()
                }
                self.experiment.log_metrics(outputs)
                results.append(outputs)
                # on_step_end()

            dic = self.train_end(results)        # pylint: disable=E1111

            if len(dic) != 0:
                self._metrics.update(dic)
                self.experiment.log_metrics(dic, step=epoch)

    def val_cycle(self, epoch: int, loader: DataLoader) -> None:
        """
        ニューラルネットの評価用ループです。

        Args:
            epoch
            loader:

        Returns:

        """
        results = list()
        loader_size = len(loader)

        self.model.eval()
        with self.experiment.validate():
            with no_grad():
                for step, batch in enumerate(loader):
                    batch = send(batch, self.device)
                    # on_step_start()
                    outputs = self.val_step(batch)        # pylint: disable=E1111

                    if hasattr(self.pbar, "set_postfix"):
                        per = "{:1.0%}".format(step / loader_size)
                        self.pbar.set_postfix(
                            OrderedDict(val_batch=per), refresh=True
                        )

                    outputs = {
                        key: outputs[key].cpu() if isinstance(outputs[key], Tensor) else outputs[key]
                        for key in outputs.keys()
                    }
                    self.experiment.log_metrics(outputs)
                    results.append(outputs)
                    # on_step_end()

                dic = self.val_end(results)        # pylint: disable=E1111

                if len(dic) != 0:
                    self._metrics.update(dic)
                    self.experiment.log_metrics(dic, step=epoch)

    def test_cycle(self, loader: DataLoader) -> None:
        """
        ニューラルネットの検証用ループです。

        Args:
            loader:

        Returns:

        """
        results = list()
        loader_size = len(loader)

        self.model.eval()
        with self.experiment.test():
            with no_grad():
                for step, batch in enumerate(loader):
                    batch = send(batch, self.device)
                    # on_step_start()
                    outputs = self.test_step(batch)        # pylint: disable=E1111

                    per = "{:1.0%}".format(step / loader_size)
                    if hasattr(self.pbar, "set_postfix"):
                        self.pbar.set_postfix(
                            OrderedDict(test_batch=per), refresh=True
                        )
                        self.pbar.update(1)

                    outputs = {
                        key: outputs[key].cpu() if isinstance(outputs[key], Tensor) else outputs[key]
                        for key in outputs.keys()
                    }

                    self.experiment.log_metrics(outputs)
                    results.append(outputs)
                    # on_step_end()

                dic = self.test_end(results)        # pylint: disable=E1111

                if len(dic) != 0:
                    self._metrics.update(dic)
                    self.experiment.log_metrics(dic)

    def train_config(self, epochs: int, checkpoint_path: Optional[str] = None, monitor: Optional[str] = None):
        """
        .run() メソッドを用いて訓練を行う際に、 epochs などを指定する為のメソッドです。

        Examples:
            >>> runner: BaseRunner = ...
            >>> runner.train_config(
            >>>     epochs=10,
            >>>     checkpoint_path="/path/to/checkpoint_dir",
            >>>     monitor="validate_avg_acc >= 0.75"
            >>> )

        Args:
            epochs (int): 訓練epoch数を指定する。
            checkpoint_path: checkpointを保存するディレクトリ名を指定。monitorを指定しなければ、全てのepochにおける重みを保存する。
            monitor: 指定した式に当てはまるepochのみを保存する。checkpoint_pathと一緒に設定する必要があるます。

        Returns:
            None

        """

        self.configures["checkpoint_path"]: str = checkpoint_path
        if monitor:
            mode = re.search("train|validate", monitor)
            trigger = re.split("train\w|validate\w", monitor)[1]
            self.configures["monitor"] = {"mode": mode, "trigger": trigger}

        if epochs > 0:
            self.configures["epochs"] = epochs
        else:
            self.configures["epochs"] = 1

        return self

    def log_hyperparams(self, dic: Dict = None, prefix: Optional[str] = None) -> None:
        """

        Args:
            dic (Dict):
            prefix (str):

        Returns:
            None

        """
        self.experiment.log_parameters(self.optimizer.__dict__["defaults"], prefix="optimizer")
        self.experiment.log_parameter("Optimizer", self.optimizer.__class__.__name__)

        if dic is not None:
            self.experiment.log_parameters(dic, prefix)

    def initialize(self) -> None:
        """
        Runner の準備を行うメソッドです。
        self.model, self.optimizer, self.experiment といった実行に必要な最低限の変数が未定義な場合はエラーメッセージを出し終了します。
        問題がなければ CPU or GPU にモデルを渡します。

        Returns:
            None

        """

        if not isinstance(self.model, Module):
            raise NotImplementedError("`self.model` is not defined.")

        if not isinstance(self.optimizer, Optimizer):
            raise NotImplementedError("`self.optimizer` is not defined.")

        if self.experiment is NotImplemented:
            raise NotImplementedError("`self.experiment` is not defined.")

        if self.scheduler and not isinstance(self.scheduler, list):
            raise ValueError("`scheduler` must be a list object.")

        self.model = self.model.to(self.device)

    def run(self, phase: str = "all", verbose: bool = True):
        """
        Runnerを実行します。
        実行には、事前に self.add_loader("train", train_loader) で訓練用のデータローダを登録するしておく必要があります。

        Args:
            phase:
                - `train`
                - `val`
                - `test`
                - `all`
                - `debug`
                のいずれかを指定してする事で、実行フェーズを決定します。デフォルト: all

            verbose: True の場合、学習の進行を表示します。

        Returns:
            None

        """
        phases = {"train", "train/val", "test", "all", "debug"}
        if phase not in phases:
            raise KeyError("The argument 'phase' must be one of the following. {}".format(phases))

        if phase == "debug":
            self.experiment.add_tag("debug")

        self.initialize()
        self.log_hyperparams()

        if not self.loaders:
            raise Exception("At least one DataLoader must be provided.")

        if phase in {"all", "train", "train/val", "debug"}:
            if "train" in self.loaders:
                self.pbar = tqdm(range(self.configures["epochs"]), desc="Epochs") if verbose \
                    else range(self.configures["epochs"])
                # .on_epoch_start()
                for epoch in self.pbar:
                    # on_train_start()
                    self.train_cycle(epoch, self.loaders["train"])
                    # on_train_end()

                    if phase in {"all", "train/val", "debug"}:
                        if "val" in self.loaders:
                            # on_validation_start()
                            self.val_cycle(epoch, self.loaders["val"])
                            # on_validation_end()

                    if self.scheduler:
                        self.update_scheduler(epoch)

                    if self.early_stop:
                        if self.early_stop.on_epoch_end(self._metrics, epoch):
                            break
                        # .on_epoch_end()

                    if self.configures["checkpoint_path"]:
                        ops = {
                            "==": operator.eq, "!=": operator.ne, "<": operator.lt,
                            "<=": operator.le, ">": operator.gt, ">=": operator.ge
                        }

                        def _save(o, k, v, e):
                            # o: Operation, k: Key, v: Value, e: Epoch
                            if ops[o](self.experiment.get_metric(k), v):
                                super().save(self.configures["checkpoint_path"], epoch=e)

                        if "monitor" in self.configures.keys():
                            mode = self.configures["monitor"]["mode"]
                            key, op, value = self.configures["monitor"]["trigger"].split(" ")

                            if mode == "train":
                                with self.experiment.train():
                                    _save(op, key, value, epoch)
                            elif mode == "validate":
                                with self.experiment.validate():
                                    _save(op, key, value, epoch)
                            else:
                                pass

                        super().save(self.configures["checkpoint_path"], epoch=epoch)

        if phase in {"all", "test", "debug"}:
            if "test" in self.loaders:
                # on_test_start()
                self.pbar = tqdm(total=len(self.loaders["test"]), desc="Evaluating") if verbose else None
                self.test_cycle(self.loaders["test"])
                # on_test_end()

        return self

    def predict(self, x: Union[Tensor, ndarray]) -> ndarray:
        """
        与えられた入力をもとに予測を行うメソッドです。

        Args:
            x (Union[torch.Tensor, np.ndarray]):

        Returns:
            predict
        """
        raise NotImplementedError

    def add_loader(self, mode: str, loader: DataLoader):
        """
        訓練等に用いるデータローダをRunnerに登録する為のメソッドです。

        Args:
            mode (str): ['train', 'val', 'test'] のいずれをか指定します。
            loader (torch.utils.data.DataLoader):

        Examples:
            >>> loader = DataLoader(...)
            >>> runner: BaseRunner = ...
            >>> runner.add_loader("train", loader)

        """
        if mode not in ["train", "val", "test"]:
            raise KeyError("argument `mode` must be one of 'train', 'val', or 'test'.")

        if not isinstance(loader, DataLoader):
            raise TypeError("The argument `loader` must be an instance of `torch.utils.data.DataLoader`.")

        self._loaders[mode] = loader
        return self

    @property
    def loaders(self) -> Dict[str, DataLoader]:
        return self._loaders

    def fit(self, x: ndarray, y: ndarray, **kwargs):
        """
        Scikit-Learn スタイルの訓練メソッドです。

        Args:
            x: 訓練用データ
            y: 教師ラベル
            **kwargs:

        """
        val_size: float = kwargs.get("val_size", 0.1)
        num_workers = kwargs.get("num_workers", 0)
        batch_size = kwargs.get("batch_size", 1)
        pin_memory = kwargs.get("pin_memory", False)
        verbose = kwargs.get("verbose", True)
        checkpoint_path = kwargs.get("checkpoint_path", None)
        monitor = kwargs.get("monitor", None)

        if self.configures["epochs"] == 0:
            epochs = kwargs.get("epochs", 1)
        else:
            epochs = self.configures["epochs"]

        train_ds = get_dataset(x, y)
        val_ds = get_dataset(x, y)
        n_train = len(train_ds)
        indices = list(range(n_train))
        split = int(floor(val_size * n_train))

        train_idx, val_idx = indices[split:], indices[:split]
        train_sampler = SubsetRandomSampler(train_idx)
        val_sampler = SubsetRandomSampler(val_idx)

        train_loader = DataLoader(
            train_ds, batch_size=batch_size,
            sampler=train_sampler, num_workers=num_workers,
            pin_memory=pin_memory
        )
        val_loader = DataLoader(
            val_ds, batch_size=batch_size,
            sampler=val_sampler, num_workers=num_workers, pin_memory=pin_memory
        )

        self.add_loader("train", train_loader)
        self.add_loader("val", val_loader)
        self.train_config(epochs, checkpoint_path=checkpoint_path, monitor=monitor)
        self.run(verbose=verbose)

        return self

    def freeze(self) -> None:
        """
        モデルのパラメータが勾配を計算しないように固定する為のメソッドです。

        """
        for param in self.model.parameters():
            param.requires_grad = False

        self.model.eval()

    def unfreeze(self) -> None:
        """
        .freeze() で固定されたパラメータを再度学習できるようにする為のメソッドです。

        """
        for param in self.model.parameters():
            param.requires_grad = True

        self.model.train()

    def save(self, directory: Optional[str] = None, epoch: Optional[int] = None) -> None:
        super(BaseRunner, self).save(directory, epoch)



import torch
import torch.jit
import numpy as np
import enchanter.addons as addons


def test_swish1():
    swish = torch.jit.script(addons.Swish())
    x = torch.tensor([1.0], dtype=torch.float32)
    out = swish(x)
    out = np.round(out.detach().numpy(), 4)
    assert out == np.array([0.7311]).astype(np.float32)


def test_swish2():
    swish = torch.jit.script(addons.Swish(beta=True))
    x = torch.tensor([1.0], dtype=torch.float32)
    out = swish(x)
    out = np.round(out.detach().numpy(), 4)
    assert out == np.array([0.7311]).astype(np.float32)


def test_mish1():
    mish = torch.jit.script(addons.Mish())
    x = torch.tensor([1.0], dtype=torch.float32)
    out = mish(x)
    out = np.round(out.detach().numpy(), 4)
    assert out == np.array([0.8651]).astype(np.float32)


def test_mish2():
    x = torch.tensor([1.0], dtype=torch.float32)
    out = addons.mish(x)
    out = np.round(out.detach().numpy(), 4)
    assert out == np.array([0.8651]).astype(np.float32)



# ***************************************************
#  _____            _                 _
# | ____|_ __   ___| |__   __ _ _ __ | |_ ___ _ __
# |  _| | '_ \ / __| '_ \ / _` | '_ \| __/ _ \ '__|
# | |___| | | | (__| | | | (_| | | | | ||  __/ |
# |_____|_| |_|\___|_| |_|\__,_|_| |_|\__\___|_|
#
# ***************************************************

import numpy as np
from torch import tensor, bmm
from torch.tensor import Tensor
from torch.nn import Module


__all__ = [
    "DenseInterpolation"
]


class DenseInterpolation(Module):
    """

    Args:
        seq_len: length of input sequence.
        factor:

    """
    def __init__(self, seq_len: int, factor: int) -> None:
        super(DenseInterpolation, self).__init__()
        W = np.zeros((factor, seq_len), dtype=np.float32)

        for t in range(seq_len):
            s = np.array((factor * (t + 1)) / seq_len, dtype=np.float32)
            for m in range(factor):
                tmp = np.array(1 - (np.abs(s - (1 + m)) / factor), dtype=np.float32)
                w = np.power(tmp, 2, dtype=np.float32)
                W[m, t] = w

        W = tensor(W).float().unsqueeze(0)
        self.register_buffer("W", W)

    def forward(self, x: Tensor) -> Tensor:
        """
        Dense Interpolation を入力に適用する。

        Args:
            x (torch.Tensor): 入力する配列の形状は、 `[N, seq_len, features]` を想定

        Returns:
            適用した結果 (torch.Tensor)
        """
        w = self.W.repeat(x.shape[0], 1, 1).requires_grad_(False)
        u = bmm(w, x)
        return u.transpose_(1, 2)



from torch import narrow, Tensor


def slice_axis(data: Tensor, axis: int, begin: int, end: int) -> Tensor:
    """

    Examples:
        >>> import torch
        >>> x = torch.tensor([
        >>>     [  1.,   2.,   3.,   4.],
        >>>     [  5.,   6.,   7.,   8.],
        >>>     [  9.,  10.,  11.,  12.]
        >>> ])
        >>>
        >>> slice_axis(x, axis=0, begin=1, end=3)
        >>> # [[  5.,   6.,   7.,   8.],
        >>> # [  9.,  10.,  11.,  12.]]
        >>>
        >>> slice_axis(x, axis=1, begin=0, end=2)
        >>> # [[  1.,   2.],
        >>> # [  5.,   6.],
        >>> # [  9.,  10.]]
        >>>
        >>> slice_axis(x, axis=1, begin=-3, end=-1)
        >>> # [[  2.,   3.],
        >>> # [  6.,   7.],
        >>> # [ 10.,  11.]]


    References:
        - `Deep Graph Library \
        <https://github.com/dmlc/dgl/blob/f25bc176d0365234ebb051d5069edff24ad2de4d/python/dgl/backend/pytorch/tensor.py#L159-L160>`_

        - `mxnet.ndarray.slice_axis \
        <https://beta.mxnet.io/api/ndarray/_autogen/mxnet.ndarray.slice_axis.html#mxnet-ndarray-slice-axis>`_

    Args:
        data: Source input
        axis: Axis along which to be sliced
        begin: The beginning index along the axis to be sliced
        end: The ending index along the axis to be sliced

    Returns:
        output - the output of this function.

    """
    if begin < 0:
        begin = data.shape[axis] + begin

    if end < 0:
        end = data.shape[axis] + end
    return narrow(data, axis, begin, end - begin)



from .classification import ClassificationRunner
from .regression import RegressionRunner
from .ensemble import HardEnsemble, SoftEnsemble, BaseEnsembleEstimator

__all__ = [
    "ClassificationRunner", "RegressionRunner",
    "HardEnsemble", "SoftEnsemble", "BaseEnsembleEstimator"
]



import torch
import torch.jit
import enchanter.addons.layers as layers


def test_1():
    x = torch.randn(1, 128, 512)
    model = layers.DenseInterpolation(128, 30)
    try:
        out = model(x)
        is_pass = True

    except Exception as e:
        print(e)
        is_pass = False

    assert is_pass


def test_2():
    x = torch.randn(1, 128, 512)
    model = torch.jit.trace(layers.DenseInterpolation(128, 30), (x,))
    try:
        out = model(x)
        is_pass = True

    except Exception as e:
        print(e)
        is_pass = False

    assert is_pass



# ***************************************************
#  _____            _                 _
# | ____|_ __   ___| |__   __ _ _ __ | |_ ___ _ __
# |  _| | '_ \ / __| '_ \ / _` | '_ \| __/ _ \ '__|
# | |___| | | | (__| | | | (_| | | | | ||  __/ |
# |_____|_| |_|\___|_| |_|\__,_|_| |_|\__\___|_|
#
# ***************************************************

import numpy as np
from typing import Dict, Any
from torch.optim.optimizer import Optimizer


__all__ = [
    "TransformerOptimizer"
]


class TransformerOptimizer:
    """
    Reference:
        `jadore801120/attention-is-all-you-need-pytorch \
        <https://github.com/jadore801120/attention-is-all-you-need-pytorch/blob/master/transformer/Optim.py>`_

    """
    def __init__(self, optimizer: Optimizer, d_model: int, warm_up: int) -> None:
        self._optimizer: Optimizer = optimizer
        self.warm_up: int = warm_up
        self.n_current_steps: int = 0
        self.init_lr: float = np.power(d_model, -0.5)

    def step(self) -> None:
        self._update_learning_rate()
        self._optimizer.step()

    def zero_grad(self) -> None:
        self._optimizer.zero_grad()

    def _get_lr_scale(self) -> np.ndarray:
        return np.min([
            np.power(self.n_current_steps, -0.5),
            np.power(self.warm_up, -1.5) * self.n_current_steps
        ])

    def get_lr(self) -> float:
        lr = self.init_lr * self._get_lr_scale()
        return lr

    def _update_learning_rate(self) -> None:
        self.n_current_steps += 1
        lr = self.get_lr()

        for param_group in self._optimizer.param_groups:
            param_group["lr"] = lr

    def state_dict(self) -> Dict[str, Any]:
        return self._optimizer.state_dict()



# ***************************************************
#  _____            _                 _
# | ____|_ __   ___| |__   __ _ _ __ | |_ ___ _ __
# |  _| | '_ \ / __| '_ \ / _` | '_ \| __/ _ \ '__|
# | |___| | | | (__| | | | (_| | | | | ||  __/ |
# |_____|_| |_|\___|_| |_|\__,_|_| |_|\__\___|_|
#
# ***************************************************

from torch.tensor import Tensor
from torch import sigmoid, tanh, tensor
from torch.nn.functional import softplus
from torch.nn import Module, Tanh, Softplus, Parameter


__all__ = [
    "Swish", "mish", "Mish"
]


class Swish(Module):
    """
    Swish活性化関数を適用します。
    """
    def __init__(self, beta: bool = False):
        super(Swish, self).__init__()
        if beta:
            self.weight = Parameter(tensor([1.]), requires_grad=True)
        else:
            self.weight = 1.0

    def forward(self, inputs: Tensor) -> Tensor:
        """
        入力値に対してSwishを適用します。

        Examples:
            >>> import torch
            >>> act = Swish()
            >>> x = torch.randn(2)
            >>> y = act(x)

        Args:
            inputs (torch.Tensor):

        Returns:
            Swishを適用した結果

        """
        out = inputs * sigmoid(self.weight * inputs)
        return out


def mish(x: Tensor) -> Tensor:
    """
    mish活性化関数を適用します。

    Examples:
        >>> import torch
        >>> x = torch.randn(2)
        >>> y = mish(y)

    Args:
        x (torch.Tensor):

    Returns:
        mishを適用した結果 (torch.Tensor)

    """
    return x * tanh(softplus(x))


class Mish(Module):
    """
    Mish活性化関数を適用します。

    """
    def __init__(self):
        super(Mish, self).__init__()
        self.tanh: Module = Tanh()
        self.softplus: Module = Softplus()

    def forward(self, inputs: Tensor) -> Tensor:
        """
        入力に対して Mish を適用します。

        Examples:
            >>> import torch
            >>> act = Mish()
            >>> x = torch.randn(2)
            >>> y = act(y)

        Args:
            inputs(torch.Tensor):

        Returns:
            mishを適用した結果 (torch.Tensor)

        """
        return inputs * self.tanh(self.softplus(inputs))



import torch
import torch.optim as optim
from enchanter.addons import TransformerOptimizer


def test_transformer_optim():
    def f(x):
        return 3 * x ** 2

    a = torch.tensor(10.0, requires_grad=True).float()
    adam = TransformerOptimizer(optim.Adam([a]), warm_up=10, d_model=100)

    is_pass = False
    try:
        for i in range(100):
            adam.zero_grad()
            out = f(a)
            out.backward()
            adam.step()
    except Exception as e:
        print(e)
        is_pass = False

    else:
        is_pass = True

    assert is_pass



import torch
import numpy as np

from enchanter.engine.modules import fix_seed
from enchanter.preprocessing.signal import transforms

fix_seed(0)

seq_len = 512
features = 10

torch_x = torch.randn(seq_len, features)      # [Seq_len, Features]
np_x = np.random.randn(seq_len, features)     # [Seq_len, Features]

torch_x_int = torch.randint(1, 10, (seq_len, features))      # [Seq_len, Features]
np_x_int = np.random.randint(1, 10, (seq_len, features))     # [Seq_len, Features]


def test_fixedwindow():
    size = 128
    is_pass = False

    try:
        _ = transforms.FixedWindow(size, -1)
    except ValueError:
        is_pass = True

    assert is_pass


def test_fixedwindow_torch_1():
    size = 128
    fw = transforms.FixedWindow(size)
    out = fw(torch_x)
    assert out.shape[0] == size
    assert out.shape[1] == features


def test_fixedwindow_torch_2():
    size = 1000
    fw = transforms.FixedWindow(size)
    is_pass = False

    try:
        _ = fw(torch_x)
    except IndexError:
        is_pass = True

    assert is_pass


def test_fixedwindow_torch_3():
    size = 128
    fw = transforms.FixedWindow(size, 1000)
    is_pass = False

    try:
        _ = fw(torch_x)
    except IndexError:
        is_pass = True

    assert is_pass


def test_fixedwindow_np_1():
    size = 128
    fw = transforms.FixedWindow(size)
    out = fw(np_x)
    assert out.shape[0] == size
    assert out.shape[1] == features


def test_fixedwindow_np_2():
    size = 1000
    fw = transforms.FixedWindow(size)
    is_pass = False

    try:
        _ = fw(np_x)
    except IndexError:
        is_pass = True

    assert is_pass


def test_fixedwindow_np_3():
    size = 128
    fw = transforms.FixedWindow(size, 1000)
    is_pass = False

    try:
        _ = fw(np_x)
    except IndexError:
        is_pass = True

    assert is_pass


def test_scaling_torch():
    start = 0.7
    end = 1.1

    scale = transforms.RandomScaling(start, end)

    y = scale(torch_x_int)

    assert (start * torch_x_int < y).sum().item() == (seq_len * features)
    assert (y < end * torch_x_int).sum().item() == (seq_len * features)


def test_scaling_np():
    start = 0.7
    end = 1.1

    scale = transforms.RandomScaling(start, end)

    y = scale(np_x_int)

    assert (start * np_x_int < y).sum().item() == (seq_len * features)
    assert (y < end * np_x_int).sum().item() == (seq_len * features)


def test_pad_1():
    pad = transforms.Pad(1000)
    y = pad(torch_x)

    assert y[512:].sum() == 0


def test_pad_2():
    pad = transforms.Pad(1000, 10.0)
    length = pad.length - seq_len
    y = pad(torch_x)
    assert y[512:].sum().item() == (10.0 * length * features)


def test_insert_compose():
    comp = transforms.Compose([
        transforms.GaussianNoise()
    ])
    comp.insert(0, transforms.FixedWindow(100))
    assert isinstance(comp.transforms[0], transforms.FixedWindow)

    comp = transforms.Compose([
        transforms.GaussianNoise()
    ])
    comp.insert(1, transforms.FixedWindow(100))
    assert isinstance(comp.transforms[1], transforms.FixedWindow)


def test_append_compose():
    comp = transforms.Compose([
        transforms.GaussianNoise()
    ])
    comp.append(transforms.FixedWindow(100))
    assert isinstance(comp.transforms[-1], transforms.FixedWindow)


def test_extend_compose():
    comp = transforms.Compose([
        transforms.GaussianNoise()
    ])
    comp.extend([
        transforms.FixedWindow(100),
        transforms.RandomScaling()
    ])
    assert isinstance(comp.transforms[0], transforms.GaussianNoise)
    assert isinstance(comp.transforms[1], transforms.FixedWindow)
    assert isinstance(comp.transforms[2], transforms.RandomScaling)



import numpy as np
from enchanter.preprocessing import signal


def test_fsw_1():
    x = np.random.randn(100, 23)
    y = np.random.randint(0, 9, 200)
    try:
        error_occur = False
        sw = signal.FixedSlidingWindow(256, overlap_rate=0.5)
        x, y = sw(x, y)
    except IndexError:
        error_occur = True

    assert error_occur


def test_fsw_2():
    x = np.random.randn(1024, 23)
    y = np.random.randint(0, 9, 1024)
    sw = signal.FixedSlidingWindow(256, overlap_rate=0.5)
    x, y = sw(x, y)
    assert isinstance(x, np.ndarray) and isinstance(y, np.ndarray)


def test_adjust_sequences_1():
    x = [
        np.random.randn(128, 5),
        np.random.randn(50, 5),
        np.random.randn(100, 5),
    ]
    max_len = max([i.shape[0] for i in x])
    new = signal.adjust_sequences(x)
    assert max_len == new.shape[1]


def test_adjust_sequences_2():
    x = [
        np.random.randn(128, 5),
        np.random.randn(50, 5),
        np.random.randn(100, 5),
    ]
    min_len = min([i.shape[0] for i in x])
    new = signal.adjust_sequences(x, np.min)
    assert min_len == new.shape[1]


def test_adjust_sequences_3():
    x = [
        np.random.randn(128, 5),
        np.random.randn(50, 5),
        np.random.randn(100, 5),
    ]

    new = signal.adjust_sequences(x, 64)
    assert new.shape[1] == 64


def test_adjust_sequences_4():
    x = [
        np.random.randn(128, 5),
        np.random.randn(50, 5),
        np.random.randn(100, 5),
    ]

    new = signal.adjust_sequences(x, dtype=np.int64)
    assert new.dtype == np.int64


def test_adjust_sequences_5():
    x = [
        np.random.randn(128, 5),
        np.random.randn(50, 5),
        np.random.randn(100, 5),
    ]

    new = signal.adjust_sequences(x, dtype=np.float64)
    assert new.dtype == np.float64


def test_adjust_sequences_6():
    x = [
        np.random.randn(128, 5),
        np.random.randn(50, 5),
        np.random.randn(100, 5),
    ]
    try:
        new = signal.adjust_sequences(x, fill=0)
        passed = True
    except Exception:
        passed = False

    assert passed


def test_adjust_sequences_7():
    x = [
        np.random.randn(128, 5),
        np.random.randn(50, 5),
        np.random.randn(100, 5),
    ]
    try:
        new = signal.adjust_sequences(x, fill=0.0)
        passed = True
    except Exception:
        passed = False

    assert passed


def test_adjust_sequences_8():
    x = [
        np.random.randn(128, 5),
        np.random.randn(50, 5),
        np.random.randn(100, 5),
    ]
    passed = False
    try:
        new = signal.adjust_sequences(x, fill="Something")
    except TypeError:
        passed = True

    assert passed



import torch
import torch.jit
import enchanter.addons.layers as layers


def test_causalconv():
    x = torch.randn(1, 6, 128, requires_grad=True)     # [N, C, L]
    model = layers.CausalConv1d(6, 32, 3)

    print(model.padding)
    out = model(x)
    out = out.sum()
    out.backward()
    assert x.grad != None



# ***************************************************
#  _____            _                 _
# | ____|_ __   ___| |__   __ _ _ __ | |_ ___ _ __
# |  _| | '_ \ / __| '_ \ / _` | '_ \| __/ _ \ '__|
# | |___| | | | (__| | | | (_| | | | | ||  __/ |
# |_____|_| |_|\___|_| |_|\__,_|_| |_|\__\___|_|
#
# ***************************************************

import math
import torch
import torch.nn as nn


__all__ = [
    "PositionalEncoding"
]


class PositionalEncoding(nn.Module):
    """
    Attention is all you need. で提案された　Transformer モデルで用いられる Positional Encoding

    References:
        `Sequence-to-Sequence Modeling with nn.Transformer and TorchText \
        <https://pytorch.org/tutorials/beginner/transformer_tutorial.html#define-the-model>`_

    """
    def __init__(self, d_model: int, seq_len: int, dropout: float = 0.1) -> None:
        """


        Args:
            d_model: the number of expected features in the encoder/decoder inputs.
            seq_len: length of input sequence.
            dropout: dropout rate.
        """
        super(PositionalEncoding, self).__init__()
        self.dropout: nn.Dropout = nn.Dropout(p=dropout)

        pe = torch.zeros(seq_len, d_model)
        position = torch.arange(0, seq_len, dtype=torch.float).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, d_model, 2).float() * (-math.log(10000.0) / d_model))
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        pe = pe.unsqueeze(0).transpose(0, 1)
        self.register_buffer("pe", pe)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """

        Args:
            x: (N, E, L)
            N ... batch size
            E ... features
            L ... seq len

        Returns:
            (N, E, L)

        """
        x = x.permute(2, 0, 1)      # [L, N, E]
        x = x + self.pe[:x.size(0), :]
        return self.dropout(x).permute(1, 2, 0)



import torch
import torch.jit
import enchanter.addons as addons


def test_mlp_1():
    x = torch.randn(32, 10)
    model = addons.layers.MLP([10, 128, 1], addons.mish)
    out = model(x)
    assert isinstance(out, torch.Tensor)


def test_mlp_2():
    x = torch.randn(32, 10)
    model = addons.layers.MLP([10, 128, 1], addons.Mish())
    jit = torch.jit.trace(model, (x,))
    out = jit(x)
    assert isinstance(out, torch.Tensor)


def test_ff_1():
    x = torch.randn(1, 128, 64)     # [N, seq_len, features]
    model = addons.layers.PositionWiseFeedForward(64)
    out = model(x)
    assert isinstance(out, torch.Tensor)


def test_ff_2():
    x = torch.randn(1, 128, 64)
    model = addons.layers.PositionWiseFeedForward(64)
    jit = torch.jit.trace(model, (x, ))
    out = jit(x)
    assert isinstance(out, torch.Tensor)



from typing import Callable, Optional

from torch import relu
from torch.tensor import Tensor
from torch.nn import Module, Conv1d

from enchanter.utils.backend import slice_axis


__all__ = [
    "CausalConv1d"
]


class CausalConv1d(Module):
    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            kernel_size: int,
            dilation: int = 1,
            activation: Optional[Callable[[Tensor], Tensor]] = relu,
            **kwargs
    ) -> None:
        """
        Causal Conv 1d

        Warnings:
            `torch.jit` を用いたJust-In-Time Compile には現在対応していません。

        Args:
            in_channels: 入力チャンネル数
            out_channels: 出力チャンネル数
            kernel_size: カーネルサイズ
            dilation: dilation
            activation: 活性化関数
            **kwargs:

        """
        super(CausalConv1d, self).__init__()
        self.dilation: int = dilation
        self.kernel_size: int = kernel_size
        self.padding: int = dilation * (kernel_size - 1)
        self.activation: Callable[[Tensor], Tensor] = activation

        self.conv1d = Conv1d(
            in_channels,
            out_channels,
            kernel_size,
            dilation=dilation,
            padding=self.padding,
            **kwargs
        )

    def forward(self, x: Tensor) -> Tensor:
        """
        順伝搬処理


        Args:
            x: [N, C, L] の３次元配列

        Returns:
            CausalConv1dを適用した結果 [N, C, L]

        """
        out = self.conv1d(x)
        if self.activation is not None:
            out = self.activation(out)

        if self.kernel_size > 0:
            out = slice_axis(out, axis=2, begin=0, end=-self.padding)

        return out



# ***************************************************
#  _____            _                 _
# | ____|_ __   ___| |__   __ _ _ __ | |_ ___ _ __
# |  _| | '_ \ / __| '_ \ / _` | '_ \| __/ _ \ '__|
# | |___| | | | (__| | | | (_| | | | | ||  __/ |
# |_____|_| |_|\___|_| |_|\__,_|_| |_|\__\___|_|
#
# ***************************************************

from collections import Counter
from typing import Union, List, Type, Optional, Callable

from numpy import array, stack, ndarray, max as np_max, zeros, nan, float32, dstack
from pandas import DataFrame
from tqdm.auto import tqdm


__all__ = [
    "FixedSlidingWindow", "adjust_sequences"
]


_Numerical = Union[int, float]

class FixedSlidingWindow:
    """Fixed sliding window.

    Examples::
        >>> import numpy as np
        >>> from enchanter.preprocessing import signal
        >>> x = np.random.randn(1024, 23)
        >>> y = np.random.randint(0, 9, 1024)
        >>> sw = signal.FixedSlidingWindow(256, overlap_rate=0.5)
        >>> x, y = sw(x, y)
        >>> x.shape     # [6, 256, 23]
        >>> y.shape     # [6, ]

    Args:
        window_size (int): Window size
        overlap_rate (float): overrap rate
        step_size (Optional[int]): step size

    Raises:
        AssertionError: an error occur when
            argument overlap_rate under 0.0 or over 1.0.n error occurred.

    """
    def __init__(self, window_size: int, overlap_rate: Union[float, None], step_size: Optional[int]=None) -> None:
        self.window_size: int = window_size

        if overlap_rate is None and step_size is not None:
            if step_size > 0:
                self.overlap = int(step_size)
        else:
            if not 0.0 < overlap_rate <= 1.0:
                raise AssertionError("overlap_rate ranges from 0.0 to 1.0")

            self.overlap = int(window_size * overlap_rate)

    def transform(self, inputs: ndarray, verbose: bool = False) -> ndarray:
        """

        Args:
            inputs: 2 or 3 dim of np.ndarray
            verbose: if True, show progress bar

        Returns:
            np.ndarray
        """
        seq_len = inputs.shape[0]
        if not seq_len > self.window_size:
            raise IndexError(
                "window size ({}) must be smaller then input sequence length ({}).".format(self.window_size, seq_len)
            )

        if verbose:
            data = []
            for i in tqdm(range(0, seq_len - self.window_size, self.overlap)):
                data.append(inputs[i:i + self.window_size])
        else:
            data = [inputs[i:i + self.window_size] for i in range(0, seq_len-self.window_size, self.overlap)]

        data = stack(data, 0)
        return data

    @staticmethod
    def clean(labels: ndarray) -> ndarray:
        """
        Clean up
        Args:
            labels:
        Returns:
        """
        tmp = []
        for lbl in labels:
            window_size = len(lbl)
            counter = Counter(lbl)
            common = counter.most_common()
            values = list(counter.values())
            if common[0][0] == 0 and values[0] == window_size // 2:
                label = common[1][0]
            else:
                label = common[0][0]

            tmp.append(label)

        return array(tmp)

    def __call__(self, data, target):
        data = self.transform(data)
        label = self.transform(target)
        label = self.clean(label)
        return data, label


def adjust_sequences(
        sequences: List[ndarray],
        max_len: Optional[Union[int, Callable[[List[int]], int]]] = None,
        fill: Union[str, _Numerical] = "ffill",
        dtype: Type = float32
) -> ndarray:
    """
    長さが一定でない系列データを一定の長さに整える関数。
    各サンプルに対して、 `max_len` よりもサンプルの系列が長い場合は、`max_len` まででそれ以降は無視され、
    `max_len` より短い場合は、足りない部分は最後の値を用いて埋められます。

    Args:
        sequences: 要素に、長さが一定でない np.ndarray オブジェクトを持つ Python配列。
                   各要素は2次元配列、第0次元目は系列の長さ、第1次元目は時系列の特徴の数で、全てのサンプルにおいて同じ特徴数である必要があります。
        max_len: 入力された全ての要素を指定した長さに加工します。
                   もし、指定されなかった場合は、与えられたサンプルの中で最も大きい系列長がmax_lenになります。
                   また、np.max や np.min, np.mean と言った関数が与えられた場合、それを用いて新しい長さの系列を生成できます。
        fill: `max_len` より短い場合は、足りない部分を埋める方法を指定します。 fill='ffill' とされた場合、最後の値を用いて埋められます。
               数値(Python int or Python float) が与えられた場合は、その値を用いて値を埋めます。
               fill=["ffill" or int or float]
        dtype:  NumPy のデータ型を指定してください。この値を元に出力系列のデータ型が決定されます。

    Examples:
        >>> import numpy as np
        >>> x = [
        >>>         np.array([[i] for i in [1, 2, 3, 4, 5]]),
        >>>         np.array([[i] for i in [1, 2, 3, 4, 5, 6, 7, 8]]),
        >>>         np.array([[i] for i in [1, 2, 3]]),
        >>> ]
        >>> out = adjust_sequences(x)
        >>> out[-1]
        >>> # array([[1.],
        >>> #        [2.],
        >>> #        [3.],
        >>> #        [3.],
        >>> #        [3.],
        >>> #        [3.],
        >>> #        [3.],
        >>> #        [3.]])
        >>> out = adjust_sequences(x, np.min)
        >>> out
        >>> # array([[[1],
        >>> #         [2],
        >>> #         [3]],
        >>> #
        >>> #        [[1],
        >>> #         [2],
        >>> #         [3]],
        >>> #
        >>> #        [[1],
        >>> #         [2],
        >>> #         [3]]])

    Returns:
        長さを調整した系列を np.dstack し、[Samples, Seq_len, Features] の3次元配列にし返します。
    """
    features = sequences[0].shape[1]

    lengths = []
    for item in sequences:
        if isinstance(item, ndarray):
            lengths.append(item.shape[0])

    if max_len is None:
        max_len = np_max(lengths)
    elif hasattr(max_len, "__call__"):
        max_len = int(max_len(lengths))

    new_seqs = []
    for seq in sequences:
        new_seq = zeros((max_len, features), dtype=dtype)
        new_seq[:, :] = nan
        new_seq = DataFrame(new_seq)

        if seq.dtype != dtype:
            seq = seq.astype(dtype)

        if max_len > seq.shape[0]:
            new_seq[:seq.shape[0]] = seq
            if fill == "ffill":
                new_seq = new_seq.ffill()
            elif isinstance(fill, int) or isinstance(fill, float):
                new_seq = new_seq.fillna(fill)
            else:
                raise TypeError
        else:
            new_seq[:max_len] = seq[:max_len]

        new_seqs.append(new_seq.values)

    return dstack(new_seqs).transpose((2, 0, 1))



import torch
from enchanter.metrics import calculate_accuracy


def test_accuracy():
    x = torch.tensor([[1, 0, 0], [1, 0, 0], [0, 1, 0], [0, 0, 1]]).float()
    y = torch.tensor([0, 1, 1, 2]).long()

    score = calculate_accuracy(x, y)
    assert score == 0.75



# ***************************************************
#  _____            _                 _
# | ____|_ __   ___| |__   __ _ _ __ | |_ ___ _ __
# |  _| | '_ \ / __| '_ \ / _` | '_ \| __/ _ \ '__|
# | |___| | | | (__| | | | (_| | | | | ||  __/ |
# |_____|_| |_|\___|_| |_|\__,_|_| |_|\__\___|_|
#
# ***************************************************

from typing import Tuple, List, Union, Optional

from sklearn.base import ClassifierMixin
from numpy import ndarray
from torch import Tensor
from torch.nn.modules import Module
from torch.nn.modules.loss import _Loss
from torch.optim.optimizer import Optimizer
from torch import no_grad, stack, tensor, as_tensor, max as torch_max
from comet_ml.experiment import BaseExperiment as BaseExperiment

from enchanter.engine import BaseRunner
from enchanter.callbacks import BaseLogger
from enchanter.callbacks import EarlyStopping
from enchanter.metrics import calculate_accuracy as calculate_accuracy


__all__ = [
    "ClassificationRunner"
]


class ClassificationRunner(BaseRunner, ClassifierMixin):
    """
    分類タスク向けの Runner です。

    Examples:
        >>> from comet_ml import Experiment
        >>> import torch
        >>> model = torch.nn.Sequential(...)
        >>> optimizer = torch.optim.Adam(model.parameters())
        >>> criterion = torch.nn.CrossEntropyLoss()
        >>> runner = ClassificationRunner(
        >>>     model,
        >>>     optimizer,
        >>>     criterion,
        >>>     Experiment()
        >>> )
        >>> runner.fit(...)
        >>> # or
        >>> runner.add_loader(...)
        >>> runner.train_config(epochs=10)
        >>> runner.run()

    """

    def __init__(
            self,
            model: Module,
            optimizer: Optimizer,
            criterion: _Loss,
            experiment: Union[BaseExperiment, BaseLogger],
            scheduler: Optional = None,
            early_stop: Optional[EarlyStopping] = None
    ) -> None:
        super(ClassificationRunner, self).__init__()
        self.model: Module = model
        self.optimizer: Optimizer = optimizer
        self.experiment: Union[BaseExperiment, BaseLogger] = experiment
        self.criterion: _Loss = criterion
        self.scheduler = scheduler
        self.early_stop = early_stop

    def train_step(self, batch: Tuple):
        x, y = batch
        out = self.model(x)
        loss = self.criterion(out, y)
        accuracy = calculate_accuracy(out, y)
        return {"loss": loss, "accuracy": accuracy}

    def train_end(self, outputs: List):
        avg_loss = stack([x["loss"] for x in outputs]).mean()
        avg_acc = stack([tensor(x["accuracy"]) for x in outputs]).mean()
        return {"avg_loss": avg_loss, "avg_acc": avg_acc}

    def val_step(self, batch: Tuple):
        return self.train_step(batch)

    def val_end(self, outputs: List):
        return self.train_end(outputs)

    def test_step(self, batch: Tuple):
        return self.train_step(batch)

    def test_end(self, outputs: List):
        return self.train_end(outputs)

    def predict(self, x: Union[Tensor, ndarray]) -> ndarray:
        self.model.eval()
        with no_grad():
            x = as_tensor(x, device=self.device)
            out = self.model(x)
            _, predicted = torch_max(out, 1)

        return predicted.cpu().numpy()



import numpy as np
import torch
import mxnet.ndarray as nd

from enchanter.utils import backend as bf


def test_slice_axis_0d():
    x = nd.array([
        [1., 2., 3., 4.],
        [5., 6., 7., 8.],
        [9., 10., 11., 12.]
    ]).astype("float32")

    ans = nd.array([
        [5., 6., 7., 8.],
        [9., 10., 11., 12.]
    ])

    mx = nd.slice_axis(x, axis=0, begin=1, end=3)
    enc = bf.slice_axis(torch.from_numpy(x.asnumpy()), axis=0, begin=1, end=3)

    assert np.sum(ans.asnumpy() == enc.numpy()) == 8
    assert np.sum(mx.asnumpy() == enc.numpy()) == 8


def test_slice_axis_1d():
    x = nd.array(
        [[1., 2., 3., 4.],
         [5., 6., 7., 8.],
         [9., 10., 11., 12.]]
    ).astype("float32")
    ans = nd.array([
        [1., 2.],
        [5., 6.],
        [9., 10.]
    ])
    mx = nd.slice_axis(x, axis=1, begin=0, end=2)
    enc = bf.slice_axis(torch.from_numpy(x.asnumpy()), axis=1, begin=0, end=2)
    assert np.sum(ans.asnumpy() == enc.numpy()) == 6
    assert np.sum(mx.asnumpy() == enc.numpy()) == 6


def test_slice_axis_2d():
    x = nd.random.randn(1, 64).astype("float32")
    mx = nd.slice_axis(x, axis=1, begin=0, end=-4).asnumpy().astype("float32")
    enc = bf.slice_axis(torch.from_numpy(x.asnumpy()), axis=1, begin=0, end=-4).numpy().astype("float32")

    assert np.sum(mx == enc) == 60


def test_slice_axis_3d():
    x = nd.random.randn(1, 6, 64).astype("float32")
    mx = nd.slice_axis(x, axis=2, begin=0, end=-4).asnumpy().astype("float32")
    enc = bf.slice_axis(torch.from_numpy(x.asnumpy()), axis=2, begin=0, end=-4).numpy().astype("float32")

    assert np.sum(mx == enc) == 360



# ***************************************************
#  _____            _                 _
# | ____|_ __   ___| |__   __ _ _ __ | |_ ___ _ __
# |  _| | '_ \ / __| '_ \ / _` | '_ \| __/ _ \ '__|
# | |___| | | | (__| | | | (_| | | | | ||  __/ |
# |_____|_| |_|\___|_| |_|\__,_|_| |_|\__\___|_|
#
# ***************************************************

from typing import List, Callable, Union, Optional
from torch import relu
from torch.tensor import Tensor
from torch.nn import Module, ModuleList, Linear, Sequential, Conv1d, ReLU


__all__ = [
    "MLP", "PositionWiseFeedForward"
]


class MLP(Module):
    """
    MLPを作成するクラス

    Args:
        shapes (List[int]): MLPの各層におけるニューロン数。int型の要素で構成される配列を想定。
                与えられる配列の第0番目の要素の値はモデルへの入力次元の数として扱われます。

        activation (Union[Callable[[torch.Tensor], torch.Tensor], nn.Module]): 活性化関数。
                    torch.relu や enchanter.addons.Mish() 等の微分可能な Callableなオブジェクト

    Examples:
        >>> import enchanter.addons as addons
        >>> model = addons.layers.MLP([10, 512, 128, 5], addons.Mish())
        >>> print(model)
        >>> # ModuleList(
        >>> #    (0): Linear(in_features=10, out_features=512, bias=True)
        >>> #    (1): Linear(in_features=512, out_features=128, bias=True)
        >>> #    (2): Linear(in_features=128, out_features=5, bias=True)
        >>> #)

    """

    def __init__(self, shapes: List[int], activation: Union[Callable[[Tensor], Tensor], Module] = relu):
        super(MLP, self).__init__()
        layers: List[Module] = []
        self.activation: Callable[[Tensor], Tensor] = activation

        for i in range(len(shapes) - 1):
            layers.append(Linear(shapes[i], shapes[i+1]))

        self.layers: Module = ModuleList(layers)

    def forward(self, x: Tensor) -> Tensor:
        for layer in self.layers[:-1]:
            x = layer(x)
            x = self.activation(x)

        x = self.layers[-1](x)
        return x



class PositionWiseFeedForward(Module):
    """
    Attention is all you need. で提案された PositionWiseFeedForward の 1×1 Conv1d を使ったバージョン


    Args:
        d_model: the number of expected features in the Position Wise Feed Forward inputs.
        expansion: Magnification rate of the number of hidden dimensions. Default: 2

    Examples:
        >>> import torch
        >>> import enchanter.addons as addons
        >>> x = torch.randn(1, 128, 512)    # [N, seq_len, features]
        >>> ff = addons.layers.PositionWiseFeedForward(512)
        >>> out = ff(x)

    """
    def __init__(self, d_model: int, expansion: int = 2) -> None:
        super(PositionWiseFeedForward, self).__init__()
        self.conv: Module = Sequential(
            Conv1d(d_model, d_model*expansion, 1),
            ReLU(),
            Conv1d(d_model*expansion, d_model, 1)
        )

    def forward(self, x: Tensor) -> Tensor:
        """
        入力に対して PositionWiseFeedForward を適用します。

        Args:
            x (torch.Tensor):

        Returns:

        """
        x = x.transpose(1, 2)
        x = self.conv(x)
        x = x.transpose(1, 2)
        return x



import torch
import torch.jit

import enchanter.addons.layers as layers


def test_se1d_1():
    x = torch.randn(1, 32, 128).float()
    model = layers.SELayer1d(32)
    try:
        out = model(x)
        is_pass = True
    except Exception:
        is_pass = False

    assert is_pass


def test_se1d_2():
    x = torch.randn(1, 32, 128).float()
    model = torch.jit.trace(layers.SELayer1d(32), (x, ))

    try:
        out = model(x)
        is_pass = True
    except Exception:
        is_pass = False

    assert is_pass


def test_se2d_1():
    x = torch.randn(1, 32, 128, 128).float()
    model = layers.SELayer2d(32)
    try:
        out = model(x)
        is_pass = True
    except Exception:
        is_pass = False

    assert is_pass


def test_se2d_2():
    x = torch.randn(1, 32, 128, 128).float()
    model = torch.jit.trace(layers.SELayer2d(32), (x, ))
    try:
        out = model(x)
        is_pass = True
    except Exception:
        is_pass = False

    assert is_pass



# ***************************************************
#  _____            _                 _
# | ____|_ __   ___| |__   __ _ _ __ | |_ ___ _ __
# |  _| | '_ \ / __| '_ \ / _` | '_ \| __/ _ \ '__|
# | |___| | | | (__| | | | (_| | | | | ||  __/ |
# |_____|_| |_|\___|_| |_|\__,_|_| |_|\__\___|_|
#
# ***************************************************

from typing import Any, Optional, Dict, Union
from contextlib import contextmanager
from torch import Tensor
from numpy import ndarray
from torch.utils.tensorboard import SummaryWriter


__all__ = [
    "BaseLogger", "TensorBoardLogger"
]


class BaseLogger:
    def __init__(self) -> None:
        self.context: Union[str, None] = None

    @contextmanager
    def train(self) -> None:
        old_state = self.context
        self.context = "train"

        yield self

        self.context = old_state

    @contextmanager
    def validate(self) -> None:
        old_state = self.context
        self.context = "validate"

        yield self

        self.context = old_state

    @contextmanager
    def test(self) -> None:
        old_state = self.context
        self.context = "test"

        yield self

        self.context = old_state

    def log_metric(
            self,
            name: str,
            value: Any,
            step: Optional[int] = None,
            epoch: Optional[int] = None,
            include_context: bool = True
    ) -> None:
        pass

    def log_metrics(
            self,
            dic: Dict,
            prefix: Optional[str] = None,
            step: Optional[int] = None,
            epoch: Optional[int] = None
    ) -> None:
        pass

    def log_parameter(self, name: str, value: Any, step: Optional[int] = None) -> None:
        pass

    def log_parameters(self, dic: Dict, prefix: Optional[str] = None, step: Optional[int] = None) -> None:
        pass

    def set_model_graph(self, *args: Any, **kwargs: Any) -> None:
        pass


class TensorBoardLogger(BaseLogger):
    def __init__(self, log_dir, *args, **kwargs):
        super(TensorBoardLogger, self).__init__()
        self.writer: SummaryWriter = SummaryWriter(log_dir, *args, **kwargs)

    def log_metric(
            self,
            name: str,
            value: Any,
            step: Optional[int] = None,
            epoch: Optional[int] = None,
            include_context: bool = True
    ) -> None:
        self.writer.add_scalar("{}/{}".format(self.context, name), value)

    def log_metrics(
            self,
            dic: Dict,
            prefix: Optional[str] = None,
            step: Optional[int] = None,
            epoch: Optional[int] = None
    ) -> None:
        for k, v in dic.items():
            self.writer.add_scalar(k, v)

    def log_parameter(self, name: str, value: Any, step: Optional[int] = None) -> None:
        if isinstance(value, Tensor) or\
                isinstance(value, ndarray) or\
                isinstance(value, float) or\
                isinstance(value, int):
            self.writer.add_scalar("{}/{}/{}".format(self.context, "params", name), value, step)

    def log_parameters(self, dic: Dict, prefix: Optional[str] = None, step: Optional[int] = None) -> Any:
        for k, v in dic.items():
            self.log_parameter(k, v, step)



from comet_ml import Optimizer
from comet_ml import exceptions
from enchanter.utils.comet import TunerConfigGenerator


def test_config_gen_1():
    is_pass = False
    try:
        config = TunerConfigGenerator(algorithm="foo")
    except ValueError:
        is_pass = True

    except Exception as e:
        print(e)

    assert is_pass


def test_config_gen_2():
    config = TunerConfigGenerator()
    config.suggest_categorical("categorical", ["a", "b", "c"])
    config.suggest_uniform("uniform", 0.0, 1.0)
    config.suggest_normal("normal", 0.0, 1.0)
    config.suggest_lognormal("lognormal", 0.0, 1.0)
    config.suggest_loguniform("loguniform", 0.0, 1.0)
    config.suggest_discrete("discrete", [10, 20, 30])

    config.export("test.config")

    is_pass = False
    try:
        opt = Optimizer("test.config", experiment_class="OfflineExperiment")

    except exceptions.OptimizerException as e:
        print(e)
        is_pass = False
    except Exception as e:
        print(e)

    else:
        is_pass = True
    assert is_pass



# *******************************************************
#  _____            _                 _
# | ____|_ __   ___| |__   __ _ _ __ | |_ ___ _ __
# |  _| | '_ \ / __| '_ \ / _` | '_ \| __/ _ \ '__|
# | |___| | | | (__| | | | (_| | | | | ||  __/ |
# |_____|_| |_|\___|_| |_|\__,_|_| |_|\__\___|_|
#
# *******************************************************

from typing import List, Union

from numpy import stack, ravel, ndarray, int as np_int
from sklearn.base import BaseEstimator
from torch import as_tensor, device, Tensor
from torch.cuda import is_available
from enchanter.engine.runner import BaseRunner

__all__ = [
    "BaseEnsembleEstimator", "SoftEnsemble", "HardEnsemble"
]


class BaseEnsembleEstimator(BaseEstimator):
    def __init__(self, runners: List[BaseRunner], mode: str = None) -> None:
        self.runners: List[BaseRunner] = runners
        self.mode: str = mode
        self.device: device = device("cuda" if is_available() else "cpu")

    def predict(self, x: Union[ndarray, Tensor]) -> List[ndarray]:
        x = as_tensor(x, device=self.device)

        predicts = []
        for runner in self.runners:
            predicts.append(runner.predict(x))

        return predicts


class SoftEnsemble(BaseEnsembleEstimator):
    """
    確率の平均をとるアンサンブル
    """

    def predict(self, x: Union[ndarray, Tensor]) -> ndarray:
        """

        Args:
            x:

        Returns:

        """
        predicts = super(SoftEnsemble, self).predict(x)

        predicts = sum(predicts)
        probs = predicts / len(self.runners)

        if self.mode == "classification":
            return probs.astype(np_int)
        else:
            return probs


class HardEnsemble(BaseEnsembleEstimator):
    """
    多数決をとるアンサンブル
    """
    def __init__(self, runners: List[BaseRunner]) -> None:
        super(HardEnsemble, self).__init__(runners, mode="classification")

    def predict(self, x: Union[ndarray, Tensor]) -> ndarray:
        """

        Args:
            x:

        Returns:

        """
        predicts = super().predict(x)
        predicts = stack(predicts)

        return ravel(predicts[0])



from torch.tensor import Tensor
from torch import nn


__all__ = [
    "SELayer1d", "SELayer2d"
]


class SELayer1d(nn.Module):
    def __init__(self, in_features: int, reduction: int = 16) -> None:
        super(SELayer1d, self).__init__()
        self.avg_pool = nn.AdaptiveAvgPool1d(1)
        self.fc = nn.Sequential(
            nn.Linear(in_features, in_features // reduction, bias=False),
            nn.ReLU(inplace=True),
            nn.Linear(in_features // reduction, in_features, bias=False),
            nn.Sigmoid()
        )

    def forward(self, x: Tensor) -> Tensor:
        b, c, _ = x.shape
        y = self.avg_pool(x).view(b, c)
        y = self.fc(y).view(b, c, 1)
        scaling = x * y.expand_as(x)

        return scaling


class SELayer2d(nn.Module):
    def __init__(self, in_features: int, reduction: int = 16) -> None:
        super(SELayer2d, self).__init__()
        self.avg_pool = nn.AdaptiveAvgPool2d(1)
        self.fc = nn.Sequential(
            nn.Linear(in_features, in_features // reduction, bias=False),
            nn.ReLU(inplace=True),
            nn.Linear(in_features // reduction, in_features, bias=False),
            nn.Sigmoid()
        )

    def forward(self, x: Tensor) -> Tensor:
        b, c, _, _ = x.shape
        y = self.avg_pool(x).view(b, c)
        y = self.fc(y).view(b, c, 1, 1)
        scaling = x * y.expand_as(x)

        return scaling



# ***************************************************
#  _____            _                 _
# | ____|_ __   ___| |__   __ _ _ __ | |_ ___ _ __
# |  _| | '_ \ / __| '_ \ / _` | '_ \| __/ _ \ '__|
# | |___| | | | (__| | | | (_| | | | | ||  __/ |
# |_____|_| |_|\___|_| |_|\__,_|_| |_|\__\___|_|
#
# ***************************************************
from typing import Tuple, List, Union, Optional

from numpy import ndarray
from sklearn.metrics import r2_score
from sklearn.base import RegressorMixin
from torch import Tensor
from torch.nn.modules import Module
from torch.nn.modules.loss import _Loss
from torch.optim.optimizer import Optimizer
from torch import stack, tensor, no_grad, as_tensor
from comet_ml.experiment import BaseExperiment as BaseExperiment

from enchanter.engine import BaseRunner
from enchanter.callbacks import EarlyStopping, BaseLogger

__all__ = [
    "RegressionRunner"
]


class RegressionRunner(BaseRunner, RegressorMixin):
    """
    回帰問題を対象にしたRunner。

    Examples:
        >>> runner = RegressionRunner(...)
        >>> runner.add_loader("train", ...)
        >>> runner.train_config(epochs=1)
        >>> runner.run()
        >>> # OR
        >>> runner = RegressionRunner(...)
        >>> runner.fit(x, y, epochs=1, batch_size=32)
    """

    def __init__(self, model: Module, optimizer: Optimizer, criterion: _Loss,
                 experiment: Union[BaseExperiment, BaseLogger], scheduler: Optional = None,
                 early_stop: Optional[EarlyStopping] = None) -> None:
        super(RegressionRunner, self).__init__()
        self.model: Module = model
        self.optimizer: Optimizer = optimizer
        self.criterion: _Loss = criterion
        self.experiment: Union[BaseExperiment, BaseLogger] = experiment
        self.scheduler = scheduler
        self.early_stop = early_stop

    def train_step(self, batch: Tuple):
        x, y = batch
        out = self.model(x)
        loss = self.criterion(out, y)
        r2 = r2_score(y.cpu().numpy(), out.cpu().detach().numpy())
        return {"loss": loss, "r2": r2}

    def train_end(self, outputs: List):
        avg_loss = stack([x["loss"] for x in outputs]).mean()
        avg_r2 = stack([tensor(x["r2"]) for x in outputs]).mean()
        return {"avg_loss": avg_loss, "avg_r2": avg_r2}

    def val_step(self, batch: Tuple):
        return self.train_step(batch)

    def val_end(self, outputs: List):
        return self.train_end(outputs)

    def test_step(self, batch: Tuple):
        return self.train_step(batch)

    def test_end(self, outputs: List):
        return self.train_end(outputs)

    def predict(self, x: Union[Tensor, ndarray]) -> ndarray:
        self.model.eval()
        with no_grad():
            x = as_tensor(x, device=self.device)
            out = self.model(x)

        return out.cpu().numpy()



# ***************************************************
#  _____            _                 _
# | ____|_ __   ___| |__   __ _ _ __ | |_ ___ _ __
# |  _| | '_ \ / __| '_ \ / _` | '_ \| __/ _ \ '__|
# | |___| | | | (__| | | | (_| | | | | ||  __/ |
# |_____|_| |_|\___|_| |_|\__,_|_| |_|\__\___|_|
#
# ***************************************************

from pprint import pformat
from random import choice, gauss, uniform
from typing import List, Callable, Union, Optional

from numpy import ndarray
from torch import from_numpy, Tensor
from torch.nn.functional import pad


class Compose:
    """
    Composes several transforms together.

    Args:
        transforms (list of ``Transform`` objects): list of transforms to compose.

    Examples:
        >>> import torch
        >>> x = torch.randn(512, 10)
        >>> transform = Compose([
        >>>     FixedWindow(128),
        >>>     GaussianNoise(),
        >>>     RandomScaling()
        >>> ])
        >>> y = transform(x)

    """
    def __init__(self, transforms: List[Callable]) -> None:
        self.transforms: List[Callable] = transforms

    def __call__(self, data: Union[Tensor, ndarray]) -> Union[Tensor, ndarray]:
        for t in self.transforms:
            data = t(data)

        return data

    def insert(self, index: int, modules: Callable) -> None:
        self.transforms.insert(index, modules)

    def append(self, module: Callable) -> None:
        self.transforms.append(module)

    def extend(self, modules: List[Callable]) -> None:
        self.transforms.extend(modules)

    def __repr__(self):
        return pformat(["({}): {}".format(i, j.__class__.__name__) for i, j in enumerate(self.transforms)], width=40)


class FixedWindow:
    """
    Cropping the input data with a fixed length window.

    Args:
        window_size (int): Length of the series to be trimmed from the input data.
        start_position (Optional[int]): Position to start clipping.

    Examples:
        >>> import numpy as np
        >>> x = np.random.randn(512, 18)    # [seq_len, features]
        >>> fw = FixedWindow(128)
        >>> out = fw(x)
        >>> out.shape       # [128, 18]

    """
    def __init__(self, window_size: int, start_position: Optional[int] = None) -> None:
        if isinstance(window_size, int):
            self.window_size: int = window_size
        else:
            raise TypeError("`window_size` must be integer.")

        if start_position:
            if start_position >= 0:
                self.start_position: Optional[int] = start_position
            else:
                raise ValueError("`start_position` must be 0 and over.")
        else:
            self.start_position: Optional[int] = start_position

    def __call__(self, data: Union[Tensor, ndarray]) -> Union[Tensor, ndarray]:
        """

        Args:
            data: input shape must be `[the length of sequence, features]`

        Returns:
            Cropped the data

        """
        seq_len = data.shape[0]

        if not seq_len > self.window_size:
            raise IndexError("`window size` must be smaller then input sequence length.")

        if not self.start_position:
            start = choice([i for i in range(seq_len - self.window_size)])       #nosec
        else:
            if (seq_len - self.window_size) >= self.start_position:
                start = self.start_position
            else:
                raise IndexError("The start position must be in the range 0 ~ (seq_len - window_size).")

        return data[start:start+self.window_size]


class GaussianNoise:
    """
    Apply gaussian noise to input data.

    Examples:
        >>> import torch
        >>> x = torch.randn(512, 10)
        >>> noise = GaussianNoise()
        >>> y = noise(x)

    Args:
        sigma: 正規分布の :math:`\sigma`
        mu: 正規分布の :math:`\mu`

    """
    def __init__(self, sigma: float = 0.01, mu: float = 0.0) -> None:
        self.noise: float = gauss(mu=mu, sigma=sigma)

    def __call__(self, data: Union[Tensor, ndarray]) -> Union[Tensor, ndarray]:
        return data + self.noise


class RandomScaling:
    """
    Scaling the original sequence by a random value in the range `start` and `end` following Eq.

    .. math::

        \mathcal{L_r}(\mathbf{ S }) = \mathbf{ S } \cdot ((\mathrm{end} - \mathrm{start}) * rand() + \mathrm{start})

    References:
        An End-to-End Multi-Task and Fusion CNN for Inertial-Based Gait Recognition

    Examples:
        >>> import torch
        >>> x = torch.randn(512, 10)
        >>> scale = RandomScaling()
        >>> y = scale(x)

    Args:
        start(float): スケーリング範囲の開始点
        end(float):　スケーリング範囲の終了点

    """
    def __init__(self, start: float = 0.7, end: float = 1.1) -> None:
        self.scale: float = ((end - start) * uniform(0.0, 1.0)) + start     #nosec

    def __call__(self, data: Union[Tensor, ndarray]) -> Union[Tensor, ndarray]:
        return data * self.scale


class Pad:
    """
    Fills the end of the given series with the specified method.

    Examples:
        >>> import torch
        >>> x = torch.randn(10, 3)  # [seq_len, features]
        >>> pad = Pad(20)
        >>> y = pad(x)              # [20, 3]
        >>> # OR
        >>> pad = Pad(20, 1.0)
        >>> y = pad(x)

    Args:
        length(int): Length of output series.
        value(Optional[float]): Value to fill.

    """
    def __init__(self, length: int, value: Optional[Union[int, float]] = None) -> None:
        self.length: int = length
        if value:
            self.value: Union[int, float] = value
        else:
            self.value: Union[int, float] = 0.0

    def __call__(self, data: Union[Tensor, ndarray]) -> Union[Tensor, ndarray]:
        from_np = False

        if isinstance(data, ndarray):
            from_np = True
            data = from_numpy(data)

        seq_len, features = data.shape
        pad_size = self.length - data.shape[0]
        if pad_size < 0:
            raise ValueError("The length of the input series is too short for the padding size.")

        data = data.reshape(1, seq_len, features)
        result = pad(data, [0, 0, 0, pad_size], value=self.value)[0]

        if from_np:
            result = result.numpy()

        return result



# ***************************************************
#  _____            _                 _
# | ____|_ __   ___| |__   __ _ _ __ | |_ ___ _ __
# |  _| | '_ \ / __| '_ \ / _` | '_ \| __/ _ \ '__|
# | |___| | | | (__| | | | (_| | | | | ||  __/ |
# |_____|_| |_|\___|_| |_|\__,_|_| |_|\__\___|_|
#
# ***************************************************

from abc import ABC
from typing import Any


__all__ = [
    "Callback"
]


class Callback(ABC):
    def on_epoch_start(self, *args: Any, **kwargs: Any) -> None:
        """Called when the epoch begins."""
        pass

    def on_epoch_end(self, *args: Any, **kwargs: Any) -> None:
        """Called when the epoch ends."""
        pass

    def on_step_start(self, *args: Any, **kwargs: Any) -> None:
        """Called when the training batch begins."""
        pass

    def on_step_end(self, *args: Any, **kwargs: Any) -> None:
        """Called when the training batch ends."""
        pass

    def on_train_start(self, *args: Any, **kwargs: Any) -> None:
        """Called when the train begins."""
        pass

    def on_train_end(self, *args: Any, **kwargs: Any) -> None:
        """Called when the train ends."""
        pass

    def on_validation_start(self, *args: Any, **kwargs: Any) -> None:
        """Called when the validation loop begins."""
        pass

    def on_validation_end(self, *args: Any, **kwargs: Any) -> None:
        """Called when the validation loop ends."""
        pass

    def on_test_start(self, *args: Any, **kwargs: Any) -> None:
        """Called when the test begins."""
        pass

    def on_test_end(self, *args: Any, **kwargs: Any) -> None:
        """Called when the test ends."""
        pass



# ***************************************************
#  _____            _                 _
# | ____|_ __   ___| |__   __ _ _ __ | |_ ___ _ __
# |  _| | '_ \ / __| '_ \ / _` | '_ \| __/ _ \ '__|
# | |___| | | | (__| | | | (_| | | | | ||  __/ |
# |_____|_| |_|\___|_| |_|\__,_|_| |_|\__\___|_|
#
# ***************************************************

from typing import Union, Tuple, Any
from os import environ as os_environ
from random import seed as std_seed
from numpy import ndarray
from numpy.random import seed as np_seed
from torch.backends import cudnn
from torch.utils.data import Dataset
from torch import device as torch_device
from torch import manual_seed, Tensor, as_tensor
from torch.cuda import is_available as cuda_is_available


from torch.utils.data import TensorDataset


__all__ = [
    "is_jupyter", "get_dataset", "fix_seed"
]


def is_jupyter() -> bool:
    """
    実行中の環境が Jupyter Notebookかどうかを判定します。

    Returns:
        True if run on jupyrer notebook else False

    """
    if "get_ipython" not in globals():
        return False
    env = get_ipython().__class__.__name__      # NOQA
    if env == "TerminalInteractiveShell":
        return False
    return True


def get_dataset(x: Union[ndarray, Tensor], y: Union[ndarray, Tensor] = None) -> Dataset:
    """
    入力された値をもとに `torch.utils.data.TensorDataset` を生成します。

    Examples:
        >>> import torch
        >>> x = torch.randn(512, 6)
        >>> y = torch.randint(0, 9, size=[512])
        >>> ds = get_dataset(x, y)

    Args:
        x (Union[np.ndarray, torch.Tensor]):
        y (Optional[Union[np.ndarray, torch.Tensor]]):

    Returns:
        `torch.utils.data.TensorDataset`
    """
    x = as_tensor(x)

    if y is not None:
        y = as_tensor(y)
        ds = TensorDataset(x, y)
    else:
        ds = TensorDataset(x)

    return ds


def send(batch: Tuple[Any], device: torch_device) -> Tuple[Any]:
    """
    Send `variable` to `device`

    Args:
        batch: Tuple which contain variable
        device: torch.device

    Returns:
        new tuple

    """

    return tuple(map(lambda x: x.to(device) if isinstance(x, Tensor) else x, batch))


def fix_seed(seed: int, deterministic: bool = False, benchmark: bool = False) -> None:
    """
    PyTorch, NumPy, Pure Python Random のSEED値を一括固定します。

    Examples:
        >>> import torch
        >>> import numpy as np
        >>> fix_seed(0)
        >>> x = torch.randn(...)
        >>> y = np.random.randn(...)

    Args:
        seed (int): SEED値
        deterministic (bool): CuDNN上で可能な限り再現性を担保するかどうか
        benchmark (bool):

    Returns:
        None
    """
    std_seed(seed)
    os_environ["PYTHONHASHSEED"] = str(seed)
    np_seed(seed)
    manual_seed(seed)

    if cuda_is_available():
        if deterministic:
            cudnn.deterministic = True
        if benchmark:
            cudnn.benchmark = False





from typing import Union, Optional, Dict
from collections import OrderedDict
from time import ctime
from pathlib import Path
from copy import deepcopy

from torch.tensor import Tensor
from torch.nn import DataParallel
from torch.serialization import save, load


__all__ = [
    "RunnerIO"
]


class RunnerIO:
    """
    PyTorch モデルの重み、Optimizerの状態といったパラメータの読み込み・保存を担当するクラス。

    """
    def __init__(self):
        self.model = NotImplemented
        self.optimizer = NotImplemented
        self.experiment = NotImplemented
        self._checkpoint_path = NotImplemented

    def save_checkpoint(self) -> Dict[str, Union[Dict[str, Tensor], dict]]:
        """
        ニューラルネットの重みと、 Optimizerの状態を辞書として出力するメソッドです。

        Returns:
            以下のキーと値を持つ辞書を返します。
                - "model_state_dict": ニューラルネットの重み
                - "optimizer_state_dict": Optimizer の状態

        """
        if isinstance(self.model, DataParallel):
            model = self.model.module.state_dict()
        else:
            model = self.model.state_dict()

        checkpoint = {
            "model_state_dict": deepcopy(model),
            "optimizer_state_dict": deepcopy(self.optimizer.state_dict())
        }
        return checkpoint

    def load_checkpoint(self, checkpoint: Dict[str, OrderedDict]):
        """
        'model_state_dict', 'optimizer_state_dict' を持つ辞書を受け取り、それらを元に モデル と Optimizer の状態を復元します。

        Args:
            checkpoint:
                以下のキーと値を持つ辞書。
                    - "model_state_dict": ニューラルネットの重み
                    - "optimizer_state_dict": Optimizer の状態
        """
        self.model.load_state_dict(checkpoint["model_state_dict"])
        self.optimizer.load_state_dict(checkpoint["optimizer_state_dict"])
        return self

    def save(self, directory: Optional[str] = None, epoch: Optional[int] = None):

        """
        指定したディレクトリにモデルとOptimizerの状態を記録したファイルを保存します。

        Args:
            directory (Optional[str]):
            epoch (Optional[int]):

        """
        if directory is None and self._checkpoint_path is not None:
            directory = self._checkpoint_path

        elif directory is None and self._checkpoint_path is None:
            raise ValueError("The argument `directory` must be specified.")

        directory = Path(directory)
        if not directory.exists():
            directory.mkdir(parents=True)
        checkpoint = self.save_checkpoint()

        if epoch is None:
            epoch = ctime().replace(" ", "_")

        filename = "checkpoint_epoch_{}.pth".format(epoch)
        path = directory / filename
        save(checkpoint, path)

        if hasattr(self.experiment, "log_asset"):
            self.experiment.log_asset(path)

    def load(self, filename: str, map_location: str = "cpu"):
        """
        指定したファイルを元にモデルとOptimizerの状態を復元します。

        Args:
            filename (str):
            map_location (str):

        """
        checkpoint = load(filename, map_location=map_location)
        self.load_checkpoint(checkpoint)

        return self



import torch
import enchanter.addons as addons


def test_pe():
    x = torch.randn(1, 128, 32).float()
    pe = addons.layers.PositionalEncoding(128, seq_len=32)
    out = pe(x)
    assert isinstance(out, torch.Tensor)



from comet_ml import OfflineExperiment

import numpy as np
import torch.nn as nn
import torch.optim as optim
from sklearn.datasets import load_iris
from sklearn.model_selection import train_test_split
from torch.utils.data import DataLoader

import enchanter.wrappers as wrappers
import enchanter.addons as addons
from enchanter.engine.modules import get_dataset


x, y = load_iris(return_X_y=True)
x_train, x_test, y_train, y_test = train_test_split(x, y, random_state=0)
x_train, x_val, y_train, y_val = train_test_split(x_train, y_train, random_state=0)

train_ds = get_dataset(x_train.astype(np.float32), y_train.astype(np.int64))
val_ds = get_dataset(x_val.astype(np.float32), y_val.astype(np.int64))
test_ds = get_dataset(x_test.astype(np.float32), y_test.astype(np.int64))

train_loader = DataLoader(train_ds, batch_size=32, shuffle=True)
val_loader = DataLoader(val_ds, batch_size=32, shuffle=False)
test_loader = DataLoader(test_ds, batch_size=32, shuffle=False)

model = addons.layers.MLP([4, 16, 3], activation=addons.Mish())
optimizer = optim.Adam(model.parameters())


def test_classification_1():
    runner = wrappers.ClassificationRunner(
        model,
        optimizer,
        nn.CrossEntropyLoss(),
        OfflineExperiment(offline_directory="../tmp")
    )
    runner.add_loader("train", train_loader).add_loader("val", val_loader).add_loader("test", test_loader)
    runner.train_config(epochs=1)

    try:
        runner.run(verbose=True)
        is_pass = True
    except Exception:
        is_pass = False

    assert is_pass is True


def test_classification_2():
    runner = wrappers.ClassificationRunner(
        model,
        optimizer,
        nn.CrossEntropyLoss(),
        OfflineExperiment(offline_directory="/tmp")
    )
    runner.train_config(epochs=1)

    try:
        runner.run(verbose=False)
        is_pass = True

    except Exception:
        is_pass = False

    assert is_pass is False


def test_classification_3():
    runner = wrappers.ClassificationRunner(
        model,
        optimizer,
        nn.CrossEntropyLoss(),
        OfflineExperiment(offline_directory="/tmp")
    )
    try:
        runner.fit(x.astype(np.float32), y.astype(np.int64))
        is_pass = True

    except Exception:
        is_pass = False

    assert is_pass is True
